/******************************************************************************
 * Copyright 2020 The siasun Transport Authors. All Rights Reserved.
 * Author: madesheng
 * Licensed under the Apache License, Version 1.0 (the "License");
 *****************************************************************************/
#ifndef COMMON_STRING_TOKENIZER_H
#define COMMON_STRING_TOKENIZER_H

#include <string>
#include <vector>

/**
 * @namespace robot::common
 * @brief robot::common
 */
namespace robot {
namespace common {

/**
 * @class StringTokenizer
 * @brief Used for splitting strings with respect to given delimiters.

StringTokenizer类主要目的是使用特定的分割符将字符串分割成多个部分。
作用与python的字符串分割函数split()类似。

默认分割符是" "

1个成员函数:
  std::string Next();分割得到的下一个子串。


1个静态成员函数:
  static std::vector<std::string> Split(const std::string &str,
                                        const std::string &delims);
使用分隔符delims对给定的字符串进行分割。分割结果可以用 Next()成员函数访问;


 */
class StringTokenizer {
 public:
  /**
   * @brief Constructor
   * @param s String to be split
   * @param delims Delimiters where the string should be split
   */
  explicit StringTokenizer(const std::string &s,
                           const std::string &delims = " ");

  /**
   * Destructor
   */
  virtual ~StringTokenizer() = default;

  /**
   * @brief Static method; creates a vector with the non-empty tokens obtained
   * from splitting.
   * @param str String to be split
   * @param delims Delimiters where the string should be split
   * @return A vector of tokens, each a substring of str
   * surrounded by delimiters.
   */
  static std::vector<std::string> Split(const std::string &str,
                                        const std::string &delims);

  /**
   * @brief The i-th time Next() is called, it returns the i-th token obtained
   * from splitting the string fed to the initializer at the given delimiters;
   * once the end is reached, will always return "".
   * @return A token, which is a substring of the string fed to the initializer
   */
  std::string Next();

 private:
  std::string s_;

  std::string delims_;

  size_t index_;

  size_t last_index_;
};



}  // namespace common
}  // namespace robot


#endif // COMMON_STRING_TOKENIZER_H
